import numpy as np, pandas as pd
from scipy import signal, integrate
from matplotlib import pyplot as plt, dates as mdates
from koolstof import infrared as ksi

filepath = "data/Humphreys_peakform/"
filepath_dbs = filepath + "Humphreys_peakform.dbs"
filepath_licor = "data/Humphreys_peakform_03.txt"
filepath_support = "data/Humphreys_peakform.xlsx"


def test_read_dbs():
    dbs = ksi.io.read_dbs(filepath_dbs)
    assert isinstance(dbs, pd.DataFrame)
    return dbs


def test_read_licor():
    licor = ksi.io.read_LI7000(filepath_licor)
    assert isinstance(licor, pd.DataFrame)
    return licor


dbs = test_read_dbs()
licor = test_read_licor()

# Get supporting metadata (flow rates)
support = pd.read_excel(filepath_support)
dbs["flow_rate"] = np.nan
dbs["sample_volume"] = np.nan
for supp_ix in support.index:
    dbs_ix = dbs[dbs.bottle == support.loc[supp_ix].bottle].index
    dbs.loc[dbs_ix, "flow_rate_mlmin"] = support.loc[supp_ix].flow_rate  # in ml/min
    dbs.loc[dbs_ix, "sample_volume"] = support.loc[supp_ix].sample_volume  # in μl
dbs["flow_rate"] = dbs.flow_rate_mlmin * 1e-5 / 60  # in m3/s


def test_get_licor_samples():
    licor_with_samples = ksi.io.get_licor_samples(licor, dbs)
    assert np.all(np.isin(licor_with_samples.dbs_ix, dbs.index))
    return licor_with_samples


licor = test_get_licor_samples()


licor["x_CO2_baseline"] = ksi.baseline.als(licor.x_CO2, lam=1e9, p=1e-4, niter=10)
licor["x_CO2_corrected"] = licor.x_CO2 - licor.x_CO2_baseline
licor["flow_CO2"] = (  # in mol/m**3
    licor.x_CO2_corrected
    * licor.pressure
    * 1e-3
    / ((licor.temperature + 273.15) * 8.314_462_618)
)
peaks = {}
peaks["licor_iloc"], peaks_info = signal.find_peaks(
    licor.x_CO2_corrected, distance=60, prominence=20, width=80, threshold=0
)
peaks = pd.DataFrame(peaks)
peaks["licor_ix"] = licor.iloc[peaks.licor_iloc].index
peaks["dbs_ix"] = licor.loc[peaks.licor_ix].dbs_ix.values
peaks["prominence"] = peaks_info["prominences"]
peaks["start_ix"] = np.nan
peaks["end_ix"] = np.nan
peaks["integral"] = np.nan
for peak_ix in peaks.index:
    peak = peaks.loc[peak_ix]
    i50 = licor.x_CO2_corrected <= 0.5 * peak.prominence
    i50_left = i50 & (licor.datenum < licor.loc[peak.licor_ix].datenum)
    i50_right = i50 & (licor.datenum > licor.loc[peak.licor_ix].datenum)
    # Offsets (60 and 200) assume LI-COR file is 2 Hz & standard AIRICA settings
    start_ix = licor[i50_left].index[-1] - 60
    end_ix = licor[i50_right].index[0] + 200
    peaks.loc[peak_ix, "start_ix"] = start_ix
    peaks.loc[peak_ix, "end_ix"] = end_ix
    # Integral dx assumes LI-COR file is 2 Hz
    peaks.loc[peak_ix, "flow_integral"] = integrate.trapz(
        licor.loc[start_ix:end_ix].flow_CO2.values, dx=0.5
    )  # in mol*s/m**3
peaks["start_datetime"] = licor.loc[peaks.start_ix].datetime.values
peaks["end_datetime"] = licor.loc[peaks.end_ix].datetime.values
# Duration rounding assumes LI-COR file is 2 Hz
peaks["duration"] = (
    np.round(
        (
            (
                mdates.date2num(peaks.end_datetime)
                - mdates.date2num(peaks.start_datetime)
            )
            * (60 * 60 * 24)
        )
        * 2
    )
    / 2
)
for var in ["flow_rate", "sample_volume"]:
    peaks[var] = dbs.loc[peaks.dbs_ix, var].values
peaks["total_CO2"] = peaks.flow_integral * peaks.flow_rate  # in mol
peaks["dic"] = 1e12 * peaks.total_CO2 / peaks.sample_volume

# %%
ixs = dbs.index[-10:-7]
ix = ixs[1]

l = np.isin(licor.dbs_ix, ixs)
l2 = licor.dbs_ix == ix

fig, ax = plt.subplots(dpi=300)
licor.plot("datetime", "x_CO2", ax=ax, linewidth=1, c="k")
licor.plot("datetime", "x_CO2_corrected", ax=ax, linewidth=1, linestyle="--", c="r")

licor.loc[peaks.licor_ix].plot.scatter(
    "datetime", "x_CO2_corrected", ax=ax, c="c", zorder=100, marker="^"
)
licor.loc[peaks.start_ix].plot.scatter(
    "datetime", "x_CO2_corrected", ax=ax, c="b", zorder=100
)
licor.loc[peaks.end_ix].plot.scatter(
    "datetime", "x_CO2_corrected", ax=ax, c="lime", zorder=101, marker="+"
)
ax.axhline(0.0, c="k", linewidth=0.8)
# ax.set_xlim([np.min(licor[l2].datenum), np.max(licor[l2].datenum)])
# ax.set_ylim([-5, 10])
for i in range(len(peaks.licor_ix)):
    peaks_here = [peaks.start_ix[i], peaks.licor_ix[i], peaks.end_ix[i]]
    ax.plot(
        licor.loc[peaks_here].datetime, licor.loc[peaks_here].x_CO2_corrected, c="k"
    )



import numpy as np, pandas as pd
from scipy.optimize import least_squares
from calkulate.density import seawater_1atm_MP81
from .read import read_dbs, read_logfile


def _get_logfile_index(dbs_row, logfile):
    """[row.apply] Get index in logfile corresponding to a given row of the dbs file."""
    if dbs_row.bottle in logfile.bottle.values:
        logfile_index = np.where(
            (dbs_row.bottle == logfile.bottle)
            & (dbs_row.analysis_datetime == logfile.analysis_datetime)
        )[0]
        if np.size(logfile_index) == 1:
            logfile_index = logfile.index[logfile_index[0]]
        else:
            print(
                (
                    "{} name/date matches found between dbs and logfile @ dbs loc {}"
                ).format(np.size(logfile_index), dbs_row.name)
            )
            logfile_index = np.nan
    else:
        logfile_index = np.nan
    return logfile_index


def get_logfile_index(dbs, logfile):
    """Find the index in the logfile corresponding to each row of the dbs file and add
    this in-place to the dbs as "logfile_index".

    Parameters
    ----------
    dbs : pd.DataFrame
        The dbs file as a pandas DataFrame (imported with read_dbs).
    logfile : pd.DataFrame
        The logfile as a pandas DataFrame (imported with read_logfile).
    """
    dbs["logfile_index"] = dbs.apply(_get_logfile_index, args=[logfile], axis=1)


def _get_sample_blanks(dbs_row, logfile, use_from=6):
    """[row.apply] Calculate each sample's DIC blank value."""
    try:
        lft = logfile.loc[dbs_row.logfile_index].table
        use_minutes = lft["minutes"] >= use_from
        blank_here = lft["increments"][use_minutes].mean()
        blank_here_min = lft["increments"][use_minutes].min()
        blank_here_max = lft["increments"][use_minutes].max()
        blank_here_std = lft["increments"][use_minutes].std()
        blank_here_count = use_minutes.sum()
    except (ValueError, KeyError):
        blank_here = np.nan
        blank_here_min = np.nan
        blank_here_max = np.nan
        blank_here_std = np.nan
        blank_here_count = 0
    return pd.Series(
        {
            "blank_here": blank_here,
            "blank_here_min": blank_here_min,
            "blank_here_max": blank_here_max,
            "blank_here_std": blank_here_std,
            "blank_here_count": blank_here_count,
        }
    )


def get_sample_blanks(dbs, logfile, use_from=6):
    """Calculate each sample's DIC blank value and add this in-place to the dbs plus
    some relevant statistics.

    Parameters
    ----------
    dbs : pd.DataFrame
        The dbs file as a pandas DataFrame (imported with read_dbs).
    logfile : pd.DataFrame
        The logfile as a pandas DataFrame (imported with read_logfile).
    use_from : int, optional
        Which minute of the titration to begin counting as a blank measurement, by
        default 6.
    """
    if "logfile_index" not in dbs:
        get_logfile_index(dbs, logfile)
    dbs_blanks = dbs.apply(
        _get_sample_blanks, args=[logfile], axis=1, use_from=use_from
    )
    for blank in dbs_blanks.columns:
        dbs[blank] = dbs_blanks[blank]


def _centre_and_scale(x, x_factor=None, x_offset=None):
    if x_factor is None:
        x_factor = np.std(x)
    if x_offset is None:
        x_offset = np.mean(x)
    return (x - x_offset) / x_factor


def _de_centre_and_scale(x, x_factor, x_offset):
    return x * x_factor + x_offset


def _blank_progression(x0, datenum_scaled):
    """Estimate the changing coulometer blank during an analysis session."""
    blank = (
        x0[0]
        + x0[1] * datenum_scaled
        + x0[2] * np.exp(-(datenum_scaled - x0[3]) / x0[4])
    )
    blank = np.where(blank < 0, 0, blank)
    return blank


def _lsqfun_blank_progression(x0, datenum_scaled, blank_here):
    """Fit the changing coulometer blank during an analysis session."""
    return _blank_progression(x0, datenum_scaled) - blank_here


def _get_session_blanks(dbs_group):
    """[group.apply] Calculate blanks per analysis session."""
    x = dbs_group
    if x.blank_here.isnull().all():
        print(
            "No good blank_here values available for session '{}'".format(
                dbs_group.name
            )
        )
        blank_cols = pd.Series(
            data={
                "blank_mean": np.nan,
                "blank_median": np.nan,
                "analysis_datenum_mean": np.nan,
                "analysis_datenum_std": np.nan,
                "blank_progression": [np.nan] * 5,
                "blank_fit_std": np.nan,
                "blank_fit_rmse": np.nan,
            }
        )
    else:
        blank_here = x[x.blank_good].blank_here
        datenum_here = x[x.blank_good].analysis_datenum
        datenum_mean = datenum_here.mean()
        datenum_std = datenum_here.std()
        datenum_scaled = _centre_and_scale(
            datenum_here, x_factor=datenum_std, x_offset=datenum_mean
        )
        blank_prog = least_squares(
            _lsqfun_blank_progression,
            [30, 1, 0, 1, 1],
            args=[datenum_scaled, blank_here],
        )
        blank_cols = pd.Series(
            data={
                "blank_mean": blank_here.mean(),
                "blank_median": blank_here.median(),
                "analysis_datenum_mean": datenum_mean,
                "analysis_datenum_std": datenum_std,
                "blank_progression": blank_prog["x"],
                "blank_fit_std": np.std(blank_prog["fun"]),
                "blank_fit_rmse": np.sqrt(np.mean(blank_prog["fun"] ** 2)),
            }
        )
    return blank_cols


def get_session_blanks(dbs, logfile=None, session_col="dic_cell_id", use_from=6):
    """Calculate blanks per analysis session.

    Parameters
    ----------
    dbs : pd.DataFrame
        The dbs file as a pandas DataFrame (imported with read_dbs).
    logfile : pd.DataFrame, optional
        The logfile as a pandas DataFrame (imported with read_logfile), only necessary
        if you have not run get_sample_blanks on the dbs, by default None.
    session_col : str, optional
        The column name in the dbs that identifies analysis sessions, by default
        'dic_cell_id'.
    use_from : int, optional
        Which minute of the titration to begin counting as a blank measurement, by
        default 6.  Passed to get_sample_blanks if this has not already been run.

    Returns
    -------
    sessions : pd.DataFrame
        A table with blank fit data for each analysis session.
    """
    if "blank_here" not in dbs:
        assert (
            logfile is not None
        ), "You must either provide a logfile or run get_sample_blanks on the dbs."
        get_sample_blanks(dbs, logfile, use_from=use_from)
    if "blank_good" not in dbs:
        dbs["blank_good"] = ~dbs.blank_here.isnull()
    sessions = dbs.groupby(by=session_col).apply(_get_session_blanks)
    sessions.sort_values("analysis_datenum_mean", inplace=True)
    return sessions


def _get_counts_corrected(
    dbs,
    blank_col="blank",
    counts_col="counts",
    runtime_col="run_time",
):
    """Calculate the corrected counts."""
    return dbs[counts_col] - dbs[runtime_col] * dbs[blank_col]


def get_counts_corrected(
    dbs,
    logfile=None,
    sessions=None,
    blank_col="blank",
    counts_col="counts",
    runtime_col="run_time",
    session_col="dic_cell_id",
    use_from=6,
):
    """Determine and apply the blank corrections to get corrected counts, which are
    added to dbs in place as column 'counts_corrected'.

    Parameters
    ----------
    dbs : pd.DataFrame
        The dbs file as a pandas DataFrame (imported with read_dbs).
    logfile : pd.DataFrame, optional
        The logfile as a pandas DataFrame (imported with read_logfile), only necessary
        if you have not run get_sample_blanks on the dbs, by default None.
    sessions : pd.DataFrame, optional
        The table of analysis sessions generated by get_session_blanks, will be
        generated here if not provided, by default None.
    blank_col : str, optional
        The column name for blank values to use for corrections, by default 'blank'.
    counts_col : str, optional
        The column name for uncorrected counts, by default 'counts'.
    runtime_col : str, optional
        The column name for run time, by default 'run_time'.
    session_col : str, optional
        The column name in the dbs that identifies analysis sessions, by default
        'dic_cell_id'.
    use_from : int, optional
        Which minute of the titration to begin counting as a blank measurement, by
        default 6.  Passed to get_sample_blanks if this has not already been run.
    """
    if sessions is None:
        sessions = get_session_blanks(
            dbs, logfile=logfile, session_col=session_col, use_from=use_from
        )
    dbs["analysis_datenum_scaled"] = np.nan
    dbs["blank"] = np.nan
    for session, s in sessions.iterrows():
        l = dbs[sessions.index.name] == session
        dbs.loc[l, "analysis_datenum_scaled"] = _centre_and_scale(
            dbs.loc[l].analysis_datenum,
            x_factor=s.analysis_datenum_std,
            x_offset=s.analysis_datenum_mean,
        )
        dbs.loc[l, "blank"] = _blank_progression(
            s.blank_progression,
            dbs.loc[l].analysis_datenum_scaled,
        )
    dbs["counts_corrected"] = _get_counts_corrected(
        dbs, blank_col=blank_col, counts_col=counts_col, runtime_col=runtime_col
    )


def blank_correction(
    dbs,
    logfile,
    blank_col="blank",
    counts_col="counts",
    runtime_col="run_time",
    session_col="dic_cell_id",
    use_from=6,
):
    """Convenience wrapper for get_counts_corrected.  Returns the dbs with the blanks
    having been determined for each analysis session and counts thus corrected.

    Parameters
    ----------
    dbs : pd.DataFrame
        The dbs file as a pandas DataFrame (imported with read_dbs).
    logfile : pd.DataFrame
        The logfile as a pandas DataFrame (imported with read_logfile).
    blank_col : str, optional
        The column name for blank values to use for corrections, by default 'blank'.
    counts_col : str, optional
        The column name for uncorrected counts, by default 'counts'.
    runtime_col : str, optional
        The column name for run time, by default 'run_time'.
    session_col : str, optional
        The column name in the dbs that identifies analysis sessions, by default
        'dic_cell_id'.
    use_from : int, optional
        Which minute of the titration to begin counting as a blank measurement, by
        default 6.

    Returns
    -------
    sessions : pd.DataFrame
        A table of analysis sessions including blank correction details.
    """
    sessions = get_session_blanks(dbs, logfile=logfile, use_from=use_from)
    get_counts_corrected(dbs, sessions=sessions)
    return sessions


def get_density(dbs, temperature_analysis_dic=25.0, salinity=35.0):
    """Calculate sample densities in kg/l.

    Parameters
    ----------
    dbs : pd.DataFrame
        The dbs file.
    temperature_analysis_dic : float, optional
        Temperature of DIC analysis in degC, by default 25.0
    salinity : float, optional
        Practical salinity, by default 35.0

    Returns
    -------
    pd.DataFrame
        The dbs DataFrame with an extra column 'density_analysis_dic' containing the
        density during analysis in kg/l.
    """
    if "temperature_analysis_dic" not in dbs:
        dbs["temperature_analysis_dic"] = temperature_analysis_dic
        print(
            "dbs.temperature_analysis_dic not set; assuming {} °C.".format(
                temperature_analysis_dic
            )
        )
    if "salinity" not in dbs:
        dbs["salinity"] = salinity
        print("dbs.salinity not set; assuming {}.".format(salinity))
    dbs["density_analysis_dic"] = seawater_1atm_MP81(
        temperature=dbs.temperature_analysis_dic, salinity=dbs.salinity
    )
    return dbs


def get_standard_calibrations(dbs):
    """Calculate the calibration factor for each CRM separately and add this in-place
    to dbs as column 'k_dic_here'.

    Parameters
    ----------
    dbs : pd.DataFrame
        The dbs file as a pandas DataFrame (imported with read_dbs), having then passed
        throught blank_correction().
    sessions : pd.DataFrame
        A table of analysis sessions including blank correction details, produced by
        blank_correction().
    """
    assert "dic_certified" in dbs, "You must provide some dbs.dic_certified values."
    if "density_analysis_dic" not in dbs:
        get_density(dbs)
    dbs["k_dic_here"] = (
        dbs.dic_certified * dbs.density_analysis_dic / dbs.counts_corrected
    )


def _get_session_calibrations(dbs_group):
    """[group.apply] Calculate the session-averaged calibration factors."""
    gk = dbs_group[dbs_group.k_dic_good].k_dic_here
    return pd.Series(
        {
            "k_dic_mean": gk.mean(),
            "k_dic_std": gk.std(),
            "k_dic_count": np.sum(~np.isnan(gk)),
        }
    )


def calibrate_dic(dbs, sessions):
    """Calculate the session-averaged calibration factors and calibrate all DIC
    measurements.

    Parameters
    ----------
    dbs : pd.DataFrame
        The dbs file as a pandas DataFrame (imported with read_dbs), having then passed
        throught blank_correction().
    sessions : pd.DataFrame
        A table of analysis sessions including blank correction details, produced by
        blank_correction().
    """
    if "k_dic_here" not in dbs:
        get_standard_calibrations(dbs)
    if "k_dic_good" not in dbs:
        dbs["k_dic_good"] = ~dbs.dic_certified.isnull()
    sc = dbs.groupby(by=sessions.index.name).apply(_get_session_calibrations)
    for k, v in sc.items():
        sessions[k] = v
    dbs["k_dic"] = sessions.loc[dbs[sessions.index.name]].k_dic_mean.values
    dbs["dic"] = dbs.counts_corrected * dbs.k_dic / dbs.density_analysis_dic
    dbs["dic_offset"] = dbs.dic - dbs.dic_certified



"""Miscellaneous tools for marine carbonate chemistry and other such things."""

import string, textwrap
import numpy as np
from . import (
    airica,
    coulometer,
    crm,
    infrared,
    meta,
    molar,
    optode,
    parameterisations,
    plot,
    quaatro,
    spectro,
    vindta,
)
from .meta import __author__, __version__
from .vindta import (
    read_dbs,
    read_logfile,
    plot_session_blanks,
    plot_blanks,
    plot_k_dic,
    plot_dic_offset,
    poison_correction,
)
from .optode import pH_optode
from .plot import get_cluster_profile, cluster_profile
from .parameterisations import aou_GG92, pH_tris_DD98
from .spectro.oxygen import read_cary_oxygen


lcletter = dict(zip(range(1, 27), string.ascii_lowercase))


def sigfig(x, sf):
    """Return `x` to `sf` significant figures."""
    factor = 10.0 ** np.ceil(np.log10(np.abs(x)))
    return factor * np.around(x / factor, decimals=sf)


def hello():
    greeting = textwrap.dedent(
        r"""
        k  Miscellaneous tools for
        o  marine carbonate chemistry
        o  and other such things
        l  
        s  Version {}
        t  doi:10.5281/zenodo.3999292
        o  
        f  https://seaco2.group
        """.format(
            __version__
        )
    )
    print(greeting)



import numpy as np
import pandas as pd
from . import oxygen


def pH_NIOZ(
    absorbance_578nm, absorbance_434nm, absorbance_730nm, temperature=25, salinity=35
):
    """Calculate pH from NIOZ spectrophotometer absorbances.

    Parameters
    ----------
    absorbance_578nm : array_like
        Absorbance at 578 nm.
    absorbance_434nm : array_like
        Absorbance at 434 nm.
    absorbance_730nm : array_like
        Absorbance at 730 nm.
    temperature : array_like, optional
        Temperature in °C, by default 25.
    salinity : array_like, optional
        Practical salinity, by default 35.

    Returns
    -------
    array_like
        pH on the total scale.
    """
    WL1 = absorbance_578nm
    WL2 = absorbance_434nm
    WL3 = absorbance_730nm
    pH_total = (
        np.log10(
            (((WL1 - WL3) / (WL2 - WL3) - 0.00815 * WL1) - 0.00691)
            / (2.222 - ((WL1 - WL3) / (WL2 - WL3) - 0.00815 * WL1) * 0.1331)
        )
        + 1245.69 / (temperature + 273.15)
        + 3.8275
        + 0.00211 * (35 - salinity)
    )
    return pH_total


def pH_DSC07(
    absorbance_578nm,
    absorbance_434nm,
    absorbance_730nm,
    temperature=25,
    salinity=35,
    dye_intercept=0,
    dye_slope=1,
):
    """Calculate pH following SOP 6b, including dye addition correction if known.

    Parameters
    ----------
    absorbance_578nm : array_like
        Absorbance at 578 nm.
    absorbance_434nm : array_like
        Absorbance at 434 nm.
    absorbance_730nm : array_like
        Absorbance at 730 nm.
    temperature : array_like, optional
        Temperature in °C, by default 25.
    salinity : array_like, optional
        Practical salinity, by default 35.
    dye_intercept : array_like, optional
        Intercept of the dye correction (SOP 6b eq. 9), by default 0.
    dye_slope : array_like, optional
        Slope of the dye correction (SOP 6b eq. 9), by default 1.

    Returns
    -------
    array_like
        pH on the total scale.
    """
    WL1 = absorbance_578nm
    WL2 = absorbance_434nm
    WL3 = absorbance_730nm
    ratio_raw = (WL1 - WL3) / (WL2 - WL3)
    ratio = ratio_raw - (dye_intercept + dye_slope * ratio_raw)
    pH_total = (
        np.log10((ratio - 0.00691) / (2.222 - ratio * 0.1331))
        + 1245.69 / (temperature + 273.15)
        + 3.8275
        + 0.00211 * (35 - salinity)
    )
    return pH_total


def read_agilent_pH(filename):
    with open(filename, "rb") as f:
        lines = f.read().decode("utf-16").splitlines()

    # Get positions of data tables
    is_table = False
    table_start = []
    table_end = []
    tables = []
    for i, line in enumerate(lines):
        if line.strip().startswith("#"):
            table_start.append(i)
            is_table = True
        if is_table and line.strip() == "":
            table_end.append(i)
            is_table = False

    # Import data tables
    pH_renamer = {
        "#": "order",
        "Name": "sample_name",
        "Dilut. Factor": "dilution_factor",
        "Weight(25)": "temperature",
        "Volume(35)": "salinity",
        "pH": "pH_instrument",
        "Abs<578nm>": "abs578",
        "Abs<434nm>": "abs434",
        "Abs<730nm>": "abs730",
    }
    ts = table_start[0]
    te = table_end[0]
    pH_a = (
        pd.read_fwf(
            filename,
            encoding="utf-16",
            engine="python",
            skiprows=[*range(ts), ts + 1],
            skipfooter=len(lines) - te,
            widths=[11, 17, 15, 13, 13, 13, 14],
        )
        .rename(columns=pH_renamer)
        .set_index("order")
    )
    pH_a["sample_name"] = pH_a.sample_name.where(pH_a.sample_name.notnull(), "")
    ts = table_start[1]
    te = table_end[1]
    pH_b = (
        pd.read_fwf(
            filename,
            encoding="utf-16",
            engine="python",
            skiprows=[*range(ts), ts + 1],
            skipfooter=len(lines) - te,
            widths=[11, 17, 15, 14],
        )
        .rename(columns=pH_renamer)
        .set_index("order")
    )
    pH_b["sample_name"] = pH_b.sample_name.where(pH_b.sample_name.notnull(), "")
    for k, v in pH_b.items():
        if k == "sample_name":
            assert (pH_a.sample_name == v).all()
        else:
            pH_a[k] = v

    #  Import comments file to get non-truncated sample_name
    with open(filename.replace(".TXT", "-COMMENTS.TXT"), "rb") as f:
        lines = f.read().decode("utf-16").splitlines()

    # Get positions of data tables
    is_table = False
    table_start = []
    table_end = []
    for i, line in enumerate(lines):
        if line.strip().startswith("#"):
            table_start.append(i)
            is_table = True
        if is_table and line.strip() == "":
            table_end.append(i)
            is_table = False

    # Import middle table
    ts = table_start[1]
    te = table_end[1]
    pH_c = (
        pd.read_fwf(
            filename.replace(".TXT", "-COMMENTS.TXT"),
            encoding="utf-16",
            engine="python",
            skiprows=[*range(ts), ts + 1],
            skipfooter=len(lines) - te,
            widths=[11, 23],
        )
        .rename(columns=pH_renamer)
        .set_index("order")
    )
    pH_c["sample_name"] = pH_c.sample_name.where(pH_c.sample_name.notnull(), "")

    # Update sample_name and append
    for i, row in pH_a.iterrows():
        assert pH_c.sample_name.loc[i].startswith(row.sample_name)
    pH_a["sample_name"] = pH_c.sample_name
    pH_a["file"] = filename
    return pH_a



import numpy as np
from scipy.stats import norm
from matplotlib import pyplot as plt
from . import profiles
from .profiles import get_cluster_profile, cluster_profile
from ..meta import __version__


def add_credit(ax):
    """Add koolstof credit to figures."""
    ax.text(
        1.005,
        0,
        "koolstof v{}".format(__version__),
        alpha=0.3,
        c="k",
        fontsize=7,
        ha="left",
        va="bottom",
        rotation=-90,
        transform=ax.transAxes,
    )


def plot_duplicates(
    diffs, fmax=None, fstep=None, facecolor="xkcd:azure", linecolor="xkcd:blue"
):
    """Create a histogram of duplicate differences and calculate the corresponding
    1-sigma precision.

    Parameters
    ----------
    diffs : array-like
        The differences between the duplicate values.
    fmax : float, optional
        The maximum value to use for the histogram, by default None, in which case a
        sensible value is estimated.
    fstep : float, optional
        The step size for the histogram, by default None, in which case a sensible
        value is estimated.
    facecolor : str, optional
        The colour of the histogram bars, by default 'xkcd:azure'.
    linecolor : str, optional
        The colour of the distribution line, by default 'xkcd:blue'.

    Returns
    -------
    fig, ax : plt.figure, plt.axes
        The figure and axes created.
    """
    # Prepare variables
    diffs = np.abs(diffs[~np.isnan(diffs)])
    precision = np.mean(diffs) * np.sqrt(np.pi) / 2
    n_diffs = np.size(diffs)
    fmin = 0
    ffactor = 10 / 10 ** np.floor(np.log10(np.max(diffs)))
    if fmax is None:
        fmax = 1.2 * np.ceil(np.max(diffs) * ffactor) / ffactor
    if fstep is None:
        fstep = fmax / 10 ** np.floor(np.log10(n_diffs))
    # Draw figure
    fig, ax = plt.subplots(dpi=300)
    fx = np.linspace(fmin, fmax, num=500)
    fy = 2 * norm.pdf(fx / precision) * fstep * n_diffs / precision
    ax.hist(
        diffs,
        bins=np.arange(start=fmin, stop=fmax, step=fstep),
        facecolor=facecolor,
    )
    ax.plot(fx, fy, c=linecolor, lw=3)
    ax.set_xlabel("Absolute difference")
    ax.set_xlim((fmin, fmax))
    ax.set_ylabel("Frequency")
    ax.text(
        0.98,
        0.97,
        "1$\sigma$ = {:.2e}, $n$ = {}".format(precision, n_diffs),
        ha="right",
        va="top",
        transform=ax.transAxes,
    )
    add_credit(ax)
    fig.tight_layout()
    return fig, ax



"""Make figures to assist calibrating and QCing VINDTA datasets."""
import itertools, copy
from os import sep
import numpy as np
from matplotlib import pyplot as plt, dates as mdates
from . import get, process
from ..plot import add_credit
from ..meta import __version__


markers = itertools.cycle(("o", "^", "s", "v", "D", "<", ">"))
colours = itertools.cycle(
    (
        "xkcd:purple",
        "xkcd:green",
        "xkcd:blue",
        "xkcd:pink",
        "xkcd:brown",
        "xkcd:red",
        "xkcd:teal",
        "xkcd:orange",
    )
)


def plot_increments(
    dbs, logfile, use_from=6, ax=None, alpha=0.6, dpi=300, figsize=[6.4, 4.8], **kwargs
):
    """Plot coulometer increments by the minute, focussing on the tails.
    Any additional kwargs are passed to plt.plot().
    """
    if ax is None:
        fig, ax = plt.subplots(dpi=dpi, figsize=figsize)
    else:
        fig = ax.get_figure()
    fymax = 1.0
    for i in dbs.logfile_index:
        i_data = logfile.table[i]
        i_blank = i_data["minutes"] >= use_from
        ax.plot(
            i_data["minutes"],
            i_data["increments"],
            c="xkcd:almost black",
            alpha=0.1,
        )
        ax.scatter(
            i_data["minutes"][i_blank],
            i_data["increments"][i_blank],
            alpha=alpha,
            c="xkcd:strawberry",
            clip_on=False,
            edgecolor="none",
            s=20,
        )
        fymax = np.max([fymax, np.max(i_data["increments"][-3:])])
        not_i_blank = i_data["minutes"] < use_from
        ax.scatter(
            i_data["minutes"][not_i_blank],
            i_data["increments"][not_i_blank],
            alpha=alpha,
            c="xkcd:navy",
            clip_on=False,
            edgecolor="none",
            s=20,
        )
    ax.set_xlim([0, dbs.run_time.max() + 0.5])
    ax.set_ylim([0, fymax * 1.2])
    ax.set_xlabel("Run time / minutes")
    ax.set_ylabel("Increments / per minute")
    add_credit(ax)
    plt.tight_layout()
    return fig, ax


def plot_session_blanks(
    dbs,
    sessions,
    session,
    ax=None,
    c="xkcd:navy",
    marker="o",
    figure_path=None,
    figure_format="png",
    show_fig=True,
):
    """Draw sample blanks and their fit for one analysis session.

    Parameters
    ----------
    dbs : pd.DataFrame
        The dbs file as a pandas DataFrame (imported with read_dbs), with blank
        corrections applied.
    sessions : pd.DataFrame
        A table with blank fit data for each analysis session.
    session : str
        The name of the session to be drawn (an item from sessions.index).
    ax : matplotlib axes, optional
        A set of axes to plot onto, by default None.
    c : str, optional
        Colour for plotting, by default "xkcd:navy".
    marker : str, optional
        Marker for plotting, by default "o".
    figure_path : str, optional
        File path to save a figure, by default None, in which case it is not saved.
    figure_format : str, optional
        Format extension in which to save the figure, by default "png"
    show_fig : bool, optional
        Whether to explicitly show the figure, by default True.

    Returns
    -------
    fig, ax
        The Matplotlib figure and axes generated.
    """
    # Prepare to draw the figure
    s = sessions.loc[session]
    l = dbs[sessions.index.name] == session
    if ax is None:
        fig, ax = plt.subplots(dpi=300)
    # Create and draw fitted line
    fx = np.linspace(
        dbs[l].analysis_datenum_scaled.min(), dbs[l].analysis_datenum_scaled.max(), 500
    )
    fy = get._blank_progression(s.blank_progression, fx)
    fx = mdates.num2date(
        get._de_centre_and_scale(fx, s.analysis_datenum_std, s.analysis_datenum_mean)
    )
    ax.plot(fx, fy, c=c, label="Best fit")
    # Draw errorbars
    ax.errorbar(
        "analysis_datetime",
        "blank_here",
        yerr="blank_here_std",
        data=dbs[l & dbs.blank_good],
        alpha=0.3,
        ecolor=c,
        label=None,
        linestyle="none",
    )
    # Draw the rest of the figure
    dbs[l & dbs.blank_good].plot.scatter(
        "analysis_datetime",
        "blank_here",
        ax=ax,
        c=c,
        marker=marker,
        label="Samples used",
    )
    y_max = np.max([dbs[l & dbs.blank_good].blank_here.max(), np.max(fy)]) * 1.2
    l_ignored = l & ~dbs.blank_good & (dbs.blank_here <= y_max)
    if l_ignored.any():
        dbs[l_ignored].plot.scatter(
            "analysis_datetime",
            "blank_here",
            ax=ax,
            c="none",
            edgecolor=c,
            marker=marker,
            label="Ignored",
        )
    l_offscale = l & (dbs.blank_here > y_max)
    if l_offscale.any():
        off_x = dbs[l_offscale].analysis_datetime.values
        ax.scatter(
            off_x,
            np.full(np.size(off_x), y_max * 0.99999),
            c="none",
            edgecolor=c,
            marker="^",
            label="Off scale (ignored)",
            clip_on=False,
        )
    ax.set_ylim([0, y_max])
    ax.legend(edgecolor="k")
    ax.xaxis.set_major_locator(mdates.HourLocator())
    ax.xaxis.set_major_formatter(mdates.DateFormatter("%H"))
    ax.set_xlabel("Time of day of analysis")
    ax.set_ylabel(r"Coulometer blank / count$\cdot$minute$^{-1}$")
    ax.set_title(session)
    ax.grid(alpha=0.2)
    add_credit(ax)
    plt.tight_layout()
    if figure_path is not None:
        if not figure_path.endswith(sep):
            figure_path += sep
        plt.savefig("{}{}.{}".format(figure_path, str(session), figure_format))
    if show_fig:
        plt.show()
    return fig, ax


def plot_blanks(dbs, sessions, figure_dir=None, **kwargs):
    """Draw sample blanks and their fit for all analysis sessions.

    Parameters
    ----------
    dbs : pd.DataFrame
        The dbs file as a pandas DataFrame (imported with read_dbs), with blank
        corrections applied.
    sessions : pd.DataFrame
        A table with blank fit data for each analysis session.
    kwargs
        Passed on to plot_session_blanks().
    """
    for session in sessions.index:
        fig, ax = plot_session_blanks(dbs, sessions, session, **kwargs)
        plt.close(fig)


# def blanks(dbs, dic_sessions, ax=None, title=None, alpha=0.5, **kwargs):
#     """Plot sample-by-sample blank values.
#     Additional kwargs are passed to plt.scatter.
#     """
#     if ax is None:
#         _, ax = plt.subplots()
#     if "blank_good" not in dbs:
#         dbs["blank_good"] = True
#     dbs[~dbs.blank_good].plot.scatter(
#         "analysis_datetime",
#         "blank_here",
#         ax=ax,
#         c="xkcd:strawberry",
#         alpha=alpha,
#         **kwargs
#     )
#     dbs[dbs.blank_good].plot.scatter(
#         "analysis_datetime", "blank_here", ax=ax, c="xkcd:navy", alpha=alpha, **kwargs
#     )
#     sessions_here = dbs["cell ID"].unique()
#     for session in sessions_here:
#         sl = dbs["cell ID"] == session
#         sx_sc = process.centre_and_scale(
#             np.linspace(
#                 np.min(dbs[sl]["analysis_datenum"]),
#                 np.max(dbs[sl]["analysis_datenum"]),
#                 num=100,
#             ),
#             x_factor=dic_sessions.loc[session].analysis_datenum_std,
#             x_offset=dic_sessions.loc[session].analysis_datenum_mean,
#         )
#         sx = mdates.num2date(
#             process.de_centre_and_scale(
#                 sx_sc,
#                 dic_sessions.loc[session].analysis_datenum_std,
#                 dic_sessions.loc[session].analysis_datenum_mean,
#             )
#         )
#         ax.plot(
#             sx,
#             process.blank_progression(
#                 dic_sessions.loc[session].blank_progression, sx_sc
#             ),
#             c="xkcd:navy",
#         )
#     ax.set_xlim(
#         [
#             dbs.analysis_datetime.min() - np.timedelta64(30, "m"),
#             dbs.analysis_datetime.max() + np.timedelta64(30, "m"),
#         ]
#     )
#     ax.xaxis.set_major_locator(mdates.HourLocator())
#     ax.xaxis.set_major_formatter(mdates.DateFormatter("%H"))
#     ax.set_title(title)
#     ax.set_xlabel("Time of day")
#     ax.set_ylim([0, np.max(dbs[dbs.blank_good].blank_here) * 1.05])
#     ax.set_ylabel("Sample blank / count per minute")
#     add_credit(ax)
#     return ax


def plot_k_dic(
    dbs,
    sessions,
    ax=None,
    figure_path=None,
    figure_format="png",
    show_ignored=True,
):
    """Plot DIC calibration factors through time."""
    marker = copy.deepcopy(markers)
    colour = copy.deepcopy(colours)
    if ax is None:
        fig, ax = plt.subplots(dpi=300, figsize=(10, 6))
    for session, s in sessions.iterrows():
        m = next(marker)
        c = next(colour)
        l = dbs[sessions.index.name] == session
        l_good = l & dbs.k_dic_good
        if l_good.any():
            dbs[l_good].plot.scatter(
                "analysis_datetime",
                "k_dic_here",
                ax=ax,
                c=c,
                marker=m,
                alpha=0.7,
                label=session,
                legend=False,
            )
        if show_ignored:
            l_bad = l & ~dbs.k_dic_good & ~np.isnan(dbs.dic_certified)
            if l_bad.any():
                dbs[l_bad].plot.scatter(
                    "analysis_datetime",
                    "k_dic_here",
                    ax=ax,
                    c="none",
                    edgecolor=c,
                    marker=m,
                    legend=False,
                )
        sl = dbs[sessions.index.name] == session
        sx = np.array(
            [
                dbs.loc[sl, "analysis_datetime"].min(),
                dbs.loc[sl, "analysis_datetime"].max(),
            ]
        )
        sy = np.full_like(sx, s.k_dic_mean)
        ax.plot(sx, sy, c=c)
    # ax.legend(edgecolor="k", bbox_to_anchor=(1, 1))
    ax.set_xlabel("Analysis date and time")
    ax.set_ylabel(r"DIC calibration factor / μmol$\cdot$count$^{-1}$")
    # fac_mean = dbs.k_dic.mean()
    # fac_maxdiff = (dbs.k_dic - fac_mean).abs().max()
    # if fac_maxdiff > 0:
    #     ax.set_ylim(np.array([-1, 1]) * fac_maxdiff * 1.1 + fac_mean)
    # ax.set_xlim(
    #     [
    #         dbs.analysis_datetime.min() - np.timedelta64(30, "m"),
    #         dbs.analysis_datetime.max() + np.timedelta64(30, "m"),
    #     ]
    # )
    locator = mdates.AutoDateLocator(minticks=3, maxticks=9)
    formatter = mdates.ConciseDateFormatter(locator)
    ax.xaxis.set_major_locator(locator)
    ax.xaxis.set_major_formatter(formatter)
    ax.xaxis.get_offset_text().set_visible(False)
    ax.grid(alpha=0.2)
    add_credit(ax)
    plt.tight_layout()
    if figure_path is not None:
        if not figure_path.endswith(sep):
            figure_path += sep
        plt.savefig("{}k_dic.{}".format(figure_path, figure_format))
    return fig, ax


def plot_dic_offset(
    dbs,
    sessions,
    ax=None,
    figure_path=None,
    figure_format="png",
):
    """Plot measured minus certified DIC values through time."""
    marker = copy.deepcopy(markers)
    colour = copy.deepcopy(colours)
    if ax is None:
        fig, ax = plt.subplots(dpi=300, figsize=(10, 6))
    for session in sessions.index:
        m = next(marker)
        c = next(colour)
        l = dbs[sessions.index.name] == session
        l_good = l & dbs.k_dic_good
        if l_good.any():
            dbs[l_good].plot.scatter(
                "analysis_datetime",
                "dic_offset",
                ax=ax,
                c=c,
                marker=m,
                alpha=0.7,
                label=session,
                legend=False,
            )
        l_bad = l & ~dbs.k_dic_good & ~np.isnan(dbs.dic_offset)
        if l_bad.any():
            dbs[l_bad].plot.scatter(
                "analysis_datetime",
                "dic_offset",
                ax=ax,
                c="none",
                edgecolor=c,
                marker=m,
                legend=False,
            )
    # ax.legend(edgecolor="k", bbox_to_anchor=(1, 1))
    ax.set_xlabel("Analysis date and time")
    ax.set_ylabel(r"DIC (calibrated $-$ certified) / μmol$\cdot$kg$^{-1}$")
    ax.set_ylim(np.array([-1, 1]) * dbs[dbs.k_dic_good].dic_offset.abs().max() * 1.1)
    locator = mdates.AutoDateLocator(minticks=3, maxticks=9)
    formatter = mdates.ConciseDateFormatter(locator)
    ax.xaxis.set_major_locator(locator)
    ax.xaxis.set_major_formatter(formatter)
    ax.xaxis.get_offset_text().set_visible(False)
    ax.grid(alpha=0.2)
    ax.axhline(0, c="k", linewidth=0.8)
    add_credit(ax)
    plt.tight_layout()
    if figure_path is not None:
        if not figure_path.endswith(sep):
            figure_path += sep
        plt.savefig("{}dic_offset.{}".format(figure_path, figure_format))
    return fig, ax
